from armstrong.dev.tests.utils.backports import override_settings
from django.conf import settings
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.test.client import Client
from functools import wraps
import fudge
import os
import random

from ._utils import TestCase

from .. import constants
from .. import forms
from .. import models
from .. import views


def failed_purchase(func):
    def inner(self):
        random_text = "Some Random Text (%d)" % random.randint(1000, 2000)
        backend = self.get_backend_stub(successful=False, reason=random_text)
        self.patches = [
            fudge.patch_object(views, "backends", backend),
        ]
        fudge.clear_calls()

        data = self.random_post_data
        response = self.client.post(self.url, data)
        backend_response = backend.get_backend().purchase()["response"]
        func(self, response, random_text=random_text,
                backend_response=backend_response)
    return inner


class BaseDonationFormViewTestCase(TestCase):
    view_class = views.DonationFormView
    view_name = "donations_form"

    @property
    def url(self):
        # TODO: move this into armstrong.dev
        return reverse(self.view_name)

    def setUp(self):
        super(BaseDonationFormViewTestCase, self).setUp()
        # TODO: move this to armstrong.dev
        self.client = Client()

        # TODO: make this based off of class name and move into armstrong.dev
        settings.TEMPLATE_DIRS = (
            os.path.join(os.path.dirname(__file__), "_templates"),
        )
        self.client
        self.patches = [
            fudge.patch_object(views, "backends", self.get_backend_stub())
        ]
        fudge.clear_calls()

    def tearDown(self):
        super(BaseDonationFormViewTestCase, self).tearDown()

    def assert_in_context(self, response, name):
        # TODO: move this into armstrong.dev
        context = (response.context if hasattr(response, "context")
                else response.context_data)
        self.assertTrue(name in context,
                msg="%s was not in the context" % name)

    def assert_type_in_context(self, response, name, expected_type):
        self.assert_in_context(response, name)
        context = (response.context if hasattr(response, "context")
                else response.context_data)
        self.assertTrue(isinstance(context[name], expected_type),
                msg="%s in the context, but does not have a class of %s" % (
                        name, expected_type.__name__))

    def assert_value_in_context(self, response, name, expected_value):
        self.assert_in_context(response, name)
        context = (response.context if hasattr(response, "context")
                else response.context_data)
        self.assertEqual(context[name], expected_value,
                msg="%s in the context, but not equal to '%s'" % (
                        name, expected_value))

    def assert_template(self, template, response):
        template_names = [a.name for a in response.templates]
        self.assertTrue(template in template_names,
                msg="%s not found in templates: %s" % (
                        template, response.templates))

    def assert_form_has_errors(self, response, form_name, error_fields=None):
        self.assert_in_context(response, form_name)
        form = response.context[form_name]
        self.assertNotEqual(form.errors, [],
                msg="%s.errors was empty?" % form_name)
        if error_fields:
            for field in error_fields:
                self.assertTrue(field in form.errors,
                        msg="%s not in the errors" % field)

    def assert_subform_has_errors(self, response, subform_name,
            error_fields=None):
        form = response.context["donation_form"]
        self.assertTrue(hasattr(form, subform_name))
        subform = getattr(form, subform_name)
        if error_fields:
            for field in error_fields:
                self.assertTrue(field in subform.errors,
                        msg="%s not in the errors" % field)

    def get_view_object(self):
        view = self.view_class()
        view.request = self.factory.get(self.url)
        return view

    def get_response(self):
        response = self.client.get(self.url)
        self.assertEqual(200, response.status_code, msg="sanity check")
        return response

    def get_fake_post_request(self, confirmed=False):
        d = {} if not confirmed else {"confirmed": u"1"}
        return self.factory.post(self.url, d)

    @property
    def fake_get_request(self):
        return self.factory.get(self.url)

    def get_post_view(self, confirmed=False):
        v = views.DonationFormView(confirm=True)
        v.request = self.get_fake_post_request(confirmed=confirmed)
        return v

    post_view = property(get_post_view)


# TODO: move to armstrong.dev
def get_response(func):
    @wraps(func)
    def inner(self):
        func(self, self.get_response())
    return inner


class DonationFormViewGetTestCase(BaseDonationFormViewTestCase):
    @get_response
    def test_adds_form_action_url_to_context(self, response):
        self.assert_value_in_context(response, "form_action_url", "")

    @get_response
    def test_adds_donation_formset_to_context(self, response):
        self.assert_type_in_context(response, "donation_form",
                forms.BaseDonationForm)

    def test_get_donation_form_returns_credit_card_form_by_default(self):
        # TODO: make sure in "default" state
        view = self.get_view_object()
        donation_form = view.get_donation_form()
        self.assertIsA(donation_form, forms.CreditCardDonationForm)

    def test_get_context_turns_kwargs_into_params(self):
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }
        view = self.get_view_object()
        context = view.get_context_data(**random_kwargs)
        self.assertEqual(len(context["params"]), len(random_kwargs),
                msg="verify context.params is the same length")
        for key in context["params"].keys():
            self.assert_(key in random_kwargs)

    def test_form_is_invalid_passes_kwargs_to_get_context_data(self):
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }

        get_context_data = fudge.Fake()
        get_context_data.expects_call().with_args(**random_kwargs)

        view = self.post_view
        with fudge.patched_context(view, "get_context_data", get_context_data):
            view.post({}, **random_kwargs)

        fudge.verify()

    def test_form_is_valid_passes_kwargs_to_get_context_data(self):
        donation, donation_form = self.random_donation_and_form
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }

        get_context_data = fudge.Fake()
        get_context_data.expects_call().with_args(**random_kwargs)

        view = self.post_view
        with fudge.patched_context(view, "get_context_data", get_context_data):
            view.form_is_valid(donation_form, **random_kwargs)

        fudge.verify()

    def test_form_is_valid_passes_kwargs_to_purchase_failed(self):
        donation, donation_form = self.random_donation_and_form
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }

        view = self.post_view
        view.confirm = False
        backends = self.get_backend_stub(successful=False)
        backend_response = backends.get_backend().purchase()

        purchase_failed = fudge.Fake()
        purchase_failed.expects_call().with_args(backend_response,
                **random_kwargs)

        with fudge.patched_context(views, "backends", backends):
            with fudge.patched_context(view, "purchase_failed",
                    purchase_failed):
                view.form_is_valid(donation_form, **random_kwargs)

        fudge.verify()

    def test_purchase_failed_passes_kwargs_to_get_context_data(self):
        backend_response = {
            "reason": "Some Random Reason",
            "response": "Some Random Response",
        }
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }

        get_context_data = fudge.Fake()
        (get_context_data.expects_call()
                .with_args(**random_kwargs)
                .returns({}))

        view = self.post_view
        with fudge.patched_context(view, "get_context_data", get_context_data):
            view.purchase_failed(backend_response, **random_kwargs)

        fudge.verify()


def form_is_valid_response(confirmed=False):
    def outer(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            donation, form = self.random_donation_and_form
            fake_save = fudge.Fake().is_callable().returns(donation)
            setattr(form, "save", fake_save)
            v = self.get_post_view(confirmed=confirmed)
            response = v.form_is_valid(form)
            func(self, response)
        return inner
    return outer


class DonationFormViewPostWithConfirmTestCase(BaseDonationFormViewTestCase):
    view_name = "donations_form_confirm"

    def test_use_confirm_template_false_by_default(self):
        v = views.DonationFormView()
        self.assertFalse(v.use_confirm_template)

    def test_use_confirm_template_true_if_confirmation_required(self):
        v = self.post_view
        self.assertTrue(v.use_confirm_template)

    def test_use_confirm_template_false_if_confirmed(self):
        v = self.get_post_view(confirmed=True)
        self.assertFalse(v.use_confirm_template)

    def test_use_confirm_template_false_if_confirmation_failed(self):
        v = self.post_view
        v.form_validation_failed = True
        self.assertFalse(v.use_confirm_template)

    def test_swaps_templates_on_confirmation(self):
        v = self.post_view
        self.assertEqual(v.confirm_template_name, v.get_template_names()[0])

    def test_uses_regular_template_when_confirmation_not_required(self):
        v = self.get_post_view(confirmed=True)
        self.assertEqual(v.template_name, v.get_template_names()[0])

    def test_uses_regular_template_on_get_request(self):
        v = views.DonationFormView(confirm=True)
        v.request = self.fake_get_request
        self.assertEqual(v.template_name, v.get_template_names()[0])

    def test_uses_regular_template_on_invalid_request(self):
        v = self.post_view
        v.form_validation_failed = True
        self.assertEqual(v.template_name, v.get_template_names()[0])

    def test_form_is_invalid_uses_regular_template(self):
        v = self.post_view
        response = v.form_is_invalid()
        self.assertEqual(v.template_name, response.template_name[0])

    def test_requires_confirmation_is_true_by_default_on_posts(self):
        self.assertTrue(self.post_view.requires_confirmation)

    def test_requires_confirmation_is_false_if_confirmed(self):
        v = self.get_post_view(confirmed=True)
        self.assertFalse(v.requires_confirmation)

    @form_is_valid_response()
    def test_form_is_valid_re_renders_if_confirmation_is_required(self, r):
        self.assertIsA(r, TemplateResponse)

    @form_is_valid_response()
    def test_contains_confirmation_required_in_context(self, r):
        self.assert_value_in_context(r, "confirmation_required", True)

    @form_is_valid_response(confirmed=True)
    def test_redirects_on_confirmed(self, r):
        self.assertIsA(r, HttpResponseRedirect)

    def test_form_is_invalid_receives_kwargs_from_post(self):
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }

        donation_form = fudge.Fake()
        donation_form.provides("is_valid").returns(False)
        get_donation_form = fudge.Fake()
        get_donation_form.is_callable().returns(donation_form)

        form_is_invalid = fudge.Fake()
        form_is_invalid.expects_call().with_args(**random_kwargs)

        view = self.post_view
        with fudge.patched_context(view, "get_donation_form",
                get_donation_form):
            with fudge.patched_context(view, "form_is_invalid",
                    form_is_invalid):
                view.post({}, **random_kwargs)

        fudge.verify()

    def test_post_passes_kwargs_to_form_is_valid(self):
        r = lambda: random.randint(100, 200)
        random_kwargs = {
            "slug%d" % r(): "foo-%d" % r(),
        }
        donation_form = fudge.Fake()
        donation_form.provides("is_valid").returns(True)
        get_donation_form = fudge.Fake()
        get_donation_form.is_callable().returns(donation_form)

        form_is_valid = fudge.Fake()
        form_is_valid.expects_call().with_args(donation_form=donation_form,
                **random_kwargs)
        view = self.post_view
        with fudge.patched_context(view, "get_donation_form",
                get_donation_form):
            with fudge.patched_context(view, "form_is_valid", form_is_valid):
                view.post({}, **random_kwargs)

        fudge.verify()


class DonationFormViewPostTestCase(BaseDonationFormViewTestCase):
    @property
    def random_post_data(self):
        data = self.get_base_random_data()
        address_kwargs = self.random_address_kwargs
        prefixed_address_kwargs = self.prefix_data(address_kwargs,
                prefix="billing")
        data.update(prefixed_address_kwargs)
        return data

    def test_requires_confirmation_is_false(self):
        self.assertFalse(self.get_view_object().requires_confirmation)

    def test_saves_donation_on_post_with_minimal_information(self):
        name_kwargs = self.random_donor_kwargs
        random_amount = self.random_amount
        data = self.get_base_random_data(amount=random_amount,
                **name_kwargs)
        data.update(self.get_data_as_formset())

        # sanity check
        self.assertRaises(models.Donor.DoesNotExist,
                models.Donor.objects.get, **name_kwargs)
        with override_settings(ARMSTRONG_DONATION_FORM="SimpleDonationForm"):
            self.client.post(self.url, data)
        donor = models.Donor.objects.get(**name_kwargs)
        self.assertEqual(str(donor), " ".join(name_kwargs.values()))
        donation = models.Donation.objects.get(donor=donor)
        self.assertEqual(donation.amount, random_amount)

    def test_uses_promo_code_if_available(self):
        promo_code = self.random_discount
        name_kwargs = self.random_donor_kwargs
        random_amount = self. random_amount
        data = self.get_base_random_data(amount=random_amount,
                promo_code=promo_code.code, **name_kwargs)
        data.update(self.prefix_data(self.random_address_kwargs,
                prefix="billing"))

        self.client.post(self.url, data)
        donor = models.Donor.objects.get(**name_kwargs)
        donation = models.Donation.objects.get(donor=donor)
        self.assertEqual(promo_code, donation.code)

        d = fudge.Fake().has_attr(amount=random_amount)
        self.assertAlmostEqual(promo_code.calculate(d),
                donation.amount, places=2)

    def test_saves_address_if_present(self):
        name_kwargs = self.random_donor_kwargs
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(**name_kwargs)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))

        self.client.post(self.url, data)
        donor = models.Donor.objects.get(**name_kwargs)
        address = models.DonorAddress.objects.get(**address_kwargs)
        self.assertEqual(address, donor.address)
        self.assertEqual(address, donor.mailing_address)

    def test_saves_mailing_address_if_present(self):
        name_kwargs = self.random_donor_kwargs
        address_kwargs = self.random_address_kwargs
        mailing_address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(**name_kwargs)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        data.update(self.prefix_data(mailing_address_kwargs, prefix="mailing"))
        del data[constants.MAILING_SAME_AS_BILLING]

        self.assertEqual(0, len(models.DonorAddress.objects.all()),
            msg="sanity check")
        self.client.post(self.url, data)
        self.assertEqual(2, len(models.DonorAddress.objects.all()))
        address = models.DonorAddress.objects.get(**address_kwargs)
        mailing_address = models.DonorAddress.objects.get(
                **mailing_address_kwargs)
        self.assertNotEqual(address, mailing_address)

        donor = models.Donor.objects.get(**name_kwargs)
        self.assertEqual(address, donor.address)
        self.assertEqual(mailing_address, donor.mailing_address)

    def test_only_saves_donor_once(self):
        """
        Verify the number of queries that are run.

        This assumes that the tests are run in isolation from the backend.

        This will pass if #17594 is merged in.
        """
        data = self.random_post_data
        with self.assertNumQueries(3):
            self.client.post(self.url, data)

    def test_saves_mailing_address_if_same_as_billing_is_checked(self):
        data = self.random_post_data
        data["mailing_same_as_billing"] = u"1"
        self.client.post(self.url, data)
        donor = models.Donor.objects.get(first_name=data["first_name"],
                last_name=data["last_name"])
        self.assertEqual(donor.address, donor.mailing_address)

    def test_same_as_billing_overrides_second_address(self):
        data = self.random_post_data
        data.update(self.prefix_data(self.random_address_kwargs,
                prefix="billing"))
        data.update(self.prefix_data(self.random_address_kwargs,
                prefix="mailing"))
        data["mailing_same_as_billing"] = u"1"
        self.client.post(self.url, data)
        donor = models.Donor.objects.get(first_name=data["first_name"],
                last_name=data["last_name"])
        self.assertEqual(donor.address, donor.mailing_address)

    def test_redirects_to_success_url_after_successful_save(self):
        data = self.random_post_data
        response = self.client.post(self.url, data)
        self.assertRedirects(response, reverse("donations_thanks"))

    def test_displays_error_on_donation_form_validation_error(self):
        data = self.random_post_data
        del data["ccv_code"]
        response = self.client.post(self.url, data)
        self.assert_template("armstrong/donations/donation.html", response)
        self.assert_form_has_errors(response, "donation_form", ["ccv_code", ])

    def test_displays_errors_on_address_validation_error(self):
        data = self.random_post_data
        data["billing-address"] = ""
        response = self.client.post(self.url, data)
        self.assert_template("armstrong/donations/donation.html", response)
        self.assert_subform_has_errors(response, "billing_address_form")

    def test_displays_errors_on_mailing_address_validation_error(self):
        data = self.random_post_data
        data.update(self.prefix_data(self.random_address_kwargs,
                prefix="billing"))
        data.update(self.prefix_data(self.random_address_kwargs,
                prefix="mailing"))
        del data["mailing_same_as_billing"]
        data["mailing-address"] = ""
        response = self.client.post(self.url, data)

        self.assert_template("armstrong/donations/donation.html", response)
        self.assert_subform_has_errors(response, "mailing_address_form")

    @failed_purchase
    def test_does_redisplays_form_on_failed_donation(self, response, **kwargs):
        self.assertEqual(200, response.status_code)
        self.assert_template("armstrong/donations/donation.html", response)

    @failed_purchase
    def test_error_msg_in_context_on_failed_purchase(self, response, **kwargs):
        self.assert_value_in_context(response, "error_msg",
                "Unable to process payment")

    @failed_purchase
    def test_reason_in_context_on_failed_purchase(self, response, random_text,
            **kwargs):
        self.assert_value_in_context(response, "reason", random_text)

    @failed_purchase
    def test_response_in_context_on_failed_purchase(self, response,
            backend_response, **kwargs):
        self.assert_value_in_context(response, "response", backend_response)



from decimal import Decimal
from django.core.exceptions import ValidationError
import fudge
import random
from ._utils import generate_random_user
from ._utils import TestCase

from .. import backends
from .. import models
from ..models import (Donor, Donation, DonationType)


class DonorAddressTestCase(TestCase):
    def test_outputs_address_as_string(self):
        odd_or_even = random.randint(1, 10) % 2 is 0
        address = "%d Some St" % random.randint(100, 200)
        city = "Pleasantville" if odd_or_even else "Sometown"
        state = "TX" if odd_or_even else "CA"
        zipcode = random.randint(10000, 20000)

        donor_address = models.DonorAddress(address=address, city=city,
                state=state, zipcode=zipcode)

        expected = "%s, %s, %s, %d" % (address, city, state, zipcode)
        self.assertEqual(expected, str(donor_address))


class DonorTestCase(TestCase):
    def test_can_be_created_from_user_with_profile(self):
        user = generate_random_user()
        donor = Donor.objects.create(user=user)
        self.assertEqual(user.first_name, donor.first_name)
        self.assertEqual(user.last_name, donor.last_name)

    def test_first_name_is_given_priority_if_specified_with_user(self):
        user = generate_random_user()
        name = "Bob (%d)" % random.randint(1, 10)
        d = Donor.objects.create(user=user, first_name=name)
        self.assertEqual(name, d.first_name)
        self.assertEqual(user.last_name, d.last_name)

    def test_outputs_name_as_string(self):
        random_name = "Random %d" % random.randint(100, 200)
        random_last_name = "Name %d" % random.randint(100, 200)
        donor = models.Donor(first_name=random_name,
                last_name=random_last_name)
        self.assertEqual(" ".join([random_name, random_last_name]),
                str(donor))


class DonationTypeTestCase(TestCase):
    def test_outputs_name_as_string(self):
        random_name = "Random Name %d" % random.randint(1000, 2000)
        dt = models.DonationType(name=random_name)
        self.assertEqual(random_name, str(dt))


class DonationTypeOptionTestCase(TestCase):
    def test_repeat_default_to_zero(self):
        dt = models.DonationTypeOption.objects.create(name="Simple",
                amount=100, donation_type=self.get_base_donation_type())
        self.assertEqual(0, dt.repeat)

    def test_is_repeating_is_false_by_default(self):
        dt = models.DonationTypeOption.objects.create(name="Simple",
                amount=100, donation_type=self.get_base_donation_type())
        self.assertFalse(dt.is_repeating)

    def test_is_repeating_is_true_if_repeats_one_or_more_times(self):
        dt = models.DonationTypeOption.objects.create(name="Simple",
                amount=100, repeat=1,
                donation_type=self.get_base_donation_type())
        self.assertTrue(dt.is_repeating)

    def test_outputs_name_plus_amount_as_string(self):
        random_name = "Random Name %d" % random.randint(1000, 2000)
        random_amount = random.randint(1000, 2000)
        donation_type = models.DonationType.objects.create(name=random_name)
        option = models.DonationTypeOption(donation_type=donation_type,
                amount=random_amount)

        expected = "%s (%d)" % (random_name, random_amount)
        self.assertEqual(expected, str(option))


class DonationTestCase(TestCase):
    def test_is_repeating_is_false_by_default(self):
        d = Donation()
        self.assertFalse(d.is_repeating)

    def test_is_repeating_is_true_if_donation_type_repeats(self):
        dt = models.DonationTypeOption(amount=100, repeat=1,
                donation_type=DonationType.objects.create(name="Simple"))
        d = Donation()
        d.donation_type = dt
        self.assertTrue(d.is_repeating)

    def test_is_repeating_is_false_if_donation_type_does_not_repeat(self):
        dt = models.DonationTypeOption(amount=100,
                donation_type=DonationType.objects.create(name="Simple"))
        d = Donation()
        d.donation_type = dt
        self.assertFalse(d.is_repeating)

    def test_dispatches_to_configured_backend(self):
        m = Donation()
        random_card = "some-random-card-%d" % random.randint(1000, 2000)
        fake_backend = fudge.Fake()
        fake_backend.expects("purchase").with_args(m, random_card)
        fake_get_backend = fudge.Fake()
        fake_get_backend.is_callable().returns(fake_backend)
        with fudge.patched_context(backends, "get_backend", fake_get_backend):
            m.purchase(random_card)

        fudge.verify()

    def test_uses_donation_type_if_no_amount_provided(self):
        dt = DonationType.objects.create(name="$10")
        donation_type = models.DonationTypeOption.objects.create(
                donation_type=dt, amount="10")
        d = Donation()
        d.donation_type = donation_type
        d.donor = self.random_donor

        self.assertEqual(None, d.amount, msg="sanity check")
        d.save()
        self.assertEqual(d.amount, donation_type.amount)

    def test_code_calculate_is_called_on_save_if_present(self):
        random_amount = "%s" % self.random_amount
        d = Donation()
        d.donor = self.random_donor
        d.amount = "1000"

        code = models.PromoCode.objects.create(code="random", amount="0")

        calculate = fudge.Fake()
        calculate.expects_call().with_args(d).returns(random_amount)
        with fudge.patched_context(code, "calculate", calculate):
            d.code = code
            d.save()

        self.assertEqual(random_amount, d.amount)
        fudge.verify()

    def test_has_an_attribution_field(self):
        self.assertModelHasField(models.Donation(), "attribution",
                models.models.CharField)

    def test_has_an_anonymous_field(self):
        self.assertModelHasField(models.Donation(), "anonymous",
                models.models.BooleanField)

    def test_can_be_cast_to_string(self):
        donor = self.random_donor
        amount = self.random_amount
        donation = models.Donation(donor=donor, amount=amount)

        expected = "%s donated %s" % (donor, amount)
        self.assertEqual(expected, str(donation))


class PromoCodeTestCase(TestCase):
    def test_calculate_returns_calculated_amount(self):
        random_amount = self.random_amount
        random_discount = random.randint(10, 30)
        donation = fudge.Fake()
        donation.has_attr(amount=random_amount)

        code = models.PromoCode.objects.create(code="testing",
                amount=random_discount)
        expected = Decimal(round(
                Decimal(random_amount)
                * Decimal(1 - random_discount / 100.00), 2))
        self.assertEqual(expected, code.calculate(donation))

    def test_calculate_can_handle_amount_of_zero(self):
        random_amount = self.random_amount
        donation = fudge.Fake()
        donation.has_attr(amount=random_amount)

        code = models.PromoCode.objects.create(code="zero", amount=0)
        self.assertEqual(random_amount, code.calculate(donation))

    def test_calculate_can_handle_free_discount(self):
        donation = fudge.Fake()
        donation.has_attr(amount=100)

        code = models.PromoCode.objects.create(code="free", amount=100)
        self.assertEqual(0, code.calculate(donation))

    def test_can_handle_rounding_issues(self):
        donation = fudge.Fake().has_attr(amount=100)
        code = models.PromoCode.objects.create(code="causes issues",
                amount=13)
        self.assertEqual(Decimal("87.0"), code.calculate(donation))

    def test_can_handle_less_than_1_rounding(self):
        donation = fudge.Fake().has_attr(amount=1)
        code = models.PromoCode.objects.create(code="< $1", amount=10)
        self.assertAlmostEqual(Decimal("0.90"), code.calculate(donation))


class DonationWorkFlowTestCase(TestCase):
    def test_donations_can_be_free_form_amounts(self):
        donor = self.random_donor

        random_amount = random.randint(1, 100)
        donation = Donation.objects.create(
            amount=random_amount,
            donor=donor
        )

        self.assertEqual(random_amount, donation.amount)
        self.assertEqual(donor, donation.donor)

    def test_donations_with_a_type_use_that_information(self):
        donor = self.random_donor
        donation_type = self.random_type
        donation = Donation.objects.create(
            donation_type=donation_type,
            donor=donor
        )

        self.assertEqual(20, donation.amount)

    def test_promocodes_subtract_from_the_amount_with_a_type(self):
        donor = self.random_donor
        donation_type = self.random_type
        discount = self.random_discount
        d = Donation.objects.create(
            donation_type=donation_type,
            donor=donor,
            code=discount
        )
        self.assertEqual(discount.calculate(donation_type), d.amount)



from django.http import QueryDict
import fudge
import random
from urllib import urlencode
from ._utils import TestCase

from .. import constants
from .. import forms
from .. import models


class DjangoFormAssertionsMixin(object):
    def assert_field_error(self, form, field_name):
        self.assertFalse(form.is_valid(), msg="sanity check")
        self.assert_(field_name in form.errors,
                msg="%s not in form errors" % field_name)


class BaseDonationFormTestCase(DjangoFormAssertionsMixin, TestCase):
    def test_attribution_is_stored(self):
        random_attribution = "Random Attribution %d" % random.randint(100, 200)
        data = self.get_base_random_data()
        data["attribution"] = random_attribution
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(random_attribution, donation.attribution)

    def test_anonymous_is_off_by_default_with_no_attribution_provided(self):
        data = self.get_base_random_data()
        if "anonymous" in data:
            del data["anonymous"]
        donation = forms.BaseDonationForm(data).save()
        self.assertFalse(donation.anonymous)

    def test_anonymous_is_off_if_an_empty_anonymous_value_is_provided(self):
        data = self.get_base_random_data()
        data["anonymous"] = ""
        donation = forms.BaseDonationForm(data).save()
        self.assertFalse(donation.anonymous)

    def test_anonymous_is_checked_if_present(self):
        data = self.get_base_random_data()
        data["anonymous"] = "1"
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertTrue(donation.anonymous)

    def test_applies_promo_code(self):
        promo_code = self.random_discount
        data = self.get_base_random_data()
        data["promo_code"] = promo_code.code
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(promo_code, donation.code)

    def test_applies_promo_code_with_prefixed_form(self):
        prefix = "random%d" % random.randint(1, 9)
        promo_code = self.random_discount
        data = {
            "%s-amount" % prefix: "100",
            "%s-first_name" % prefix: "Bob",
            "%s-last_name" % prefix: "Example",
            "%s-promo_code" % prefix: promo_code.code,
        }
        form = forms.BaseDonationForm(prefix=prefix, data=data)
        donation = form.save()
        self.assertEqual(promo_code, donation.code)

    def test_can_save_with_an_empty_promo_code(self):
        data = self.get_base_random_data()
        data["promo_code"] = ""
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(None, donation.code)

    def test_does_not_save_user_by_default(self):
        data = self.get_base_random_data()
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(None, donation.donor.user)

    def test_saves_user_if_user_pk_is_submitted(self):
        user = self.random_user
        data = self.get_base_random_data()
        data["user_pk"] = user.pk
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(user.pk, donation.donor.user.pk)

    def test_behaves_if_an_empty_user_pk_is_given(self):
        data = self.get_base_random_data()
        data["user_pk"] = ""
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual(None, donation.donor.user)

    def test_saves_email_if_submitted(self):
        data = self.get_base_random_data()
        data["email"] = "bob@example.com"
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual("bob@example.com", donation.donor.email)

    def test_uses_user_email_if_no_email_is_present(self):
        user = self.random_user
        user.email = "alice@example.com"
        user.save()
        data = self.get_base_random_data()
        data["user_pk"] = user.pk
        data["email"] = ""
        form = forms.BaseDonationForm(data)
        donation = form.save()
        self.assertEqual("alice@example.com", donation.donor.email)

    def test_saves_user_if_user_pk_is_submitted_on_prefixed_form(self):
        prefix = "prefix%d" % random.randint(100, 200)
        user = self.random_user
        form = forms.BaseDonationForm(prefix=prefix, data={
            "%s-amount" % prefix: "100",
            "%s-first_name" % prefix: "Bob",
            "%s-last_name" % prefix: "Example",
            "%s-donation_type_pk" % prefix: self.random_type.pk,
            "%s-user_pk" % prefix: user.pk,
        })
        donation = form.save()
        self.assertEqual(user.pk, donation.donor.user.pk)

    def test_errors_if_more_than_two_digits_are_provided(self):
        form = forms.BaseDonationForm(data={"amount": "100.123"})
        self.assertFalse(form.is_valid(donation_only=True))
        self.assertTrue("amount" in form.errors)

    def test_donation_type_is_used_if_present(self):
        random_type = self.random_type
        form = forms.BaseDonationForm(data={
            "amount": "100",
            "first_name": "Bob",
            "last_name": "Example",
            "donation_type_pk": random_type.pk,
        })
        donation = form.save()
        self.assertEqual(random_type, donation.donation_type)

    def test_donation_type_works_with_prefixed_forms(self):
        random_type = self.random_type
        prefix = "random%d" % random.randint(1, 9)
        form = forms.BaseDonationForm(prefix=prefix, data={
            "%s-amount" % prefix: "100",
            "%s-first_name" % prefix: "Bob",
            "%s-last_name" % prefix: "Example",
            "%s-donation_type_pk" % prefix: random_type.pk,
        })
        donation = form.save()
        self.assertEqual(random_type, donation.donation_type)

    def test_billing_address_form_is_a_donoraddressform(self):
        f = forms.BaseDonationForm()
        self.assertIsA(f.billing_address_form, forms.DonorAddressForm)

    def test_mailing_address_form_is_aa_donoraddressform(self):
        f = forms.BaseDonationForm()
        self.assertIsA(f.mailing_address_form, forms.DonorAddressForm)

    def test_is_valid_uses_mailing_address_form_by_default(self):
        is_valid_true = fudge.Fake().provides("is_valid").returns(True)
        is_valid_false = fudge.Fake().provides("is_valid").returns(False)
        form = forms.BaseDonationForm(data={
                "first_name": "Foo",
                "last_name": "Bar",
                "amount": "10.00",
        })
        attrs = ["billing_address_form", "donor_form", "mailing_address_form"]
        for attr in attrs:
            setattr(form, attr, is_valid_true)
        self.assertTrue(form.is_valid())

        form.mailing_address_form = is_valid_false
        self.assertFalse(form.is_valid())

    def test_is_valid_ignores_mailing_if_same_checked(self):
        is_valid_true = fudge.Fake().provides("is_valid").returns(True)
        is_valid_false = fudge.Fake().provides("is_valid").returns(False)
        form = forms.BaseDonationForm(data={
                "first_name": "Foo",
                "last_name": "Bar",
                "amount": "10.00",
                constants.MAILING_SAME_AS_BILLING: u"1",
        })
        attrs = ["billing_address_form", "donor_form"]
        for attr in attrs:
            setattr(form, attr, is_valid_true)
        form.mailing_address_form = is_valid_false
        self.assertTrue(form.is_valid())

    def test_saves_mailing_address_if_present(self):
        name_kwargs = self.random_donor_kwargs
        address_kwargs = self.random_address_kwargs
        mailing_address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(**name_kwargs)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        data.update(self.prefix_data(mailing_address_kwargs, prefix="mailing"))
        del data[constants.MAILING_SAME_AS_BILLING]

        self.assertEqual(0, len(models.DonorAddress.objects.all()),
            msg="sanity check")
        form = forms.BaseDonationForm(data=data)
        form.save()
        self.assertEqual(2, len(models.DonorAddress.objects.all()))
        address = models.DonorAddress.objects.get(**address_kwargs)
        mailing_address = models.DonorAddress.objects.get(
                **mailing_address_kwargs)
        self.assertNotEqual(address, mailing_address)

        donor = models.Donor.objects.get(**name_kwargs)
        # import ipdb; ipdb.set_trace()
        self.assertEqual(address, donor.address)
        self.assertEqual(mailing_address, donor.mailing_address)

    def test_is_valid_is_true_if_donation_type_provided_and_no_amount(self):
        donation_type = self.random_type
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(donation_type_pk=donation_type.pk)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        del data["amount"]

        form = forms.BaseDonationForm(data=data)
        self.assertTrue(form.is_valid())

    def test_cleaned_data_is_available_with_dt_and_no_amount(self):
        donation_type = self.random_type
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(donation_type_pk=donation_type.pk)
        data.update(self.prefix_data(address_kwargs), prefix="billing")
        del data["amount"]

        form = forms.BaseDonationForm(data=data)
        form.is_valid()
        self.assertEqual(form.cleaned_data["amount"], donation_type.amount)

    def test_is_valid_returns_false_with_invalid_donation_type(self):
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(
                donation_type="unknown and unknowable")
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        del data["amount"]

        form = forms.BaseDonationForm(data=data)
        self.assertFalse(form.is_valid())

    def test_is_valid_respects_form_even_with_valid_donation_type(self):
        donation_type = self.random_type
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(donation_type_pk=donation_type.pk)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        del data["amount"]
        del data["first_name"]

        form = forms.BaseDonationForm(data=data)
        self.assertFalse(form.is_valid())

    def test_is_valid_works_with_sub_forms(self):
        donation_type = self.random_type
        address_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(donation_type_pk=donation_type.pk)
        data.update(self.prefix_data(address_kwargs, prefix="billing"))
        del data["amount"]

        class MyAwesomeForm(forms.BaseDonationForm):
            extra_field = forms.forms.BooleanField()

        form = MyAwesomeForm(data=data)
        self.assertFalse(form.is_valid())

    def test_has_a_donor_form_property(self):
        form = forms.BaseDonationForm()
        self.assertIsA(form.donor_form, forms.DonorForm)

    def test_is_valid_returns_false_if_donor_form_invalid_donor(self):
        for field_name in ["first_name", "last_name", ]:
            data = self.get_base_random_data()
            del data[field_name]
            form = forms.BaseDonationForm(data=data)
            self.assertFalse(form.is_valid())
            self.assert_field_error(form.donor_form, field_name)


class CreditCardDonationFormTestCase(TestCase):
    form_class = forms.CreditCardDonationForm

    def setUp(self):
        name_kwargs = self.random_address_kwargs
        data = self.get_base_random_data(**name_kwargs)
        self.amount = data["amount"]
        self.card_number = data["card_number"]
        self.ccv_code = data["ccv_code"]
        self.expiration_month = data["expiration_month"]
        self.expiration_year = data["expiration_year"]
        self.data = data

    def get_form(self, *args, **kwargs):
        return self.form_class(*args, **kwargs)

    def get_invalid_form(self):
        del self.data["amount"]
        return self.get_form(data=self.data)

    def assert_value_is_empty(self, field):
        form = self.get_invalid_form()
        form.data = QueryDict(urlencode(form.data))
        form.fields_to_strip = ["card_number", "ccv_code"]
        self.assertFalse(form.is_valid())
        self.assertEqual("", form[field].value())

    def assert_value_is_not_empty(self, field):
        form = self.get_invalid_form()
        form.fields_to_strip = []
        self.assertFalse(form.is_valid())
        self.assertNotEqual("", form[field].value())

    def test_clears_credit_card_numbers_if_form_is_invalid(self):
        self.assert_value_is_empty("card_number")

    def test_clears_ccv_card_numbers_if_form_is_invalid(self):
        self.assert_value_is_empty("ccv_code")

    def test_does_not_clear_card_numbers_if_form_is_invalid(self):
        self.assert_value_is_not_empty("card_number")

    def test_does_not_clear_ccv_codes_if_form_is_invalid(self):
        self.assert_value_is_not_empty("ccv_code")

    def test_can_configure_the_fields_to_strip(self):
        form = self.get_invalid_form()
        form.fields_to_strip = ["card_number", ]
        form.is_valid()

        self.assertEqual("", form["card_number"].value())
        self.assertNotEqual("", form["ccv_code"].value())


class AuthorizeDonationFormTestCase(CreditCardDonationFormTestCase):
    form_class = forms.AuthorizeDonationForm



from django.conf.urls.defaults import patterns, url
from . import views

urlpatterns = patterns('',
    url(r"^/?$", views.DonationFormView.as_view(), name="donations_form"),
    url(r"^thanks/?$", views.ThanksView.as_view(), name="donations_thanks"),
)



from armstrong.dev.tests.utils.base import ArmstrongTestCase
from armstrong.dev.tests.utils.users import generate_random_user
import datetime
from django.contrib.auth.models import User
from django.test.client import RequestFactory
import fudge
import random

from .. import forms
from .. import models
from ..models import (Donation, DonorAddress, Donor, DonationType, PromoCode)


def no_initial_patched_objects(func):
    def inner(self, *args, **kwargs):
        self.restore_patched_objects()
        self.patched = []
        return func(self, *args, **kwargs)
    return inner


class TestCase(ArmstrongTestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        # TODO: move this to armstrong.dev
        self.factory = RequestFactory()

    def tearDown(self):
        self.restore_patched_objects()

    def restore_patched_objects(self):
        if hasattr(self, "patches"):
            [p.restore() for p in self.patches]

    @property
    def random_user(self):
        return User.objects.create(
                username="random_user_%d" % random.randint(100, 200),
                first_name=self.random_donor_name,
                last_name=self.random_donor_last_name,
        )

    @property
    def random_donor_kwargs(self):
        return {
            "first_name": self.random_donor_name,
            "last_name": self.random_donor_last_name,
        }

    @property
    def random_donor_name(self):
        return "Bob (%d)" % random.randint(100, 200)

    @property
    def random_donor_last_name(self):
        return "Example (%d)" % random.randint(100, 200)

    @property
    def random_address_kwargs(self):
        return {
            "address": u"%d Some St" % random.randint(1000, 2000),
            "city": u"Anytown",
            "state": u"TX",
            "zipcode": u"787%02d" % random.randint(1, 99),
        }

    @property
    def random_address(self):
        return DonorAddress.objects.create(**self.random_address_kwargs)

    @property
    def random_donor(self):
        return Donor.objects.create(
            first_name=self.random_donor_name,
            last_name=self.random_donor_last_name,
            address=self.random_address,
            mailing_address=self.random_address
        )

    @property
    def random_donation(self):
        return Donation.objects.create(
            amount=self.random_amount,
            donor=self.random_donor,
        )

    def get_random_base_option(self, donation_type):
        return models.DonationTypeOption.objects.create(
            donation_type=donation_type,
            amount=20,
        )

    def get_base_donation_type(self):
        return models.DonationType.objects.create(name="Basic")

    @property
    def random_type(self):
        dt = self.get_base_donation_type()
        option = self.get_random_base_option(dt)
        return option

    @property
    def random_yearly_type(self):
        dt = self.get_base_donation_type()
        return models.DonationTypeOption.objects.create(
            donation_type=dt, amount=20, length=12, repeat=9999)

    @property
    def random_monthly_type(self):
        dt = self.get_base_donation_type()
        return models.DonationTypeOption.objects.create(
            donation_type=dt, amount=20, repeat=9999)

    @property
    def random_discount(self):
        r = random.randint(10, 30)
        return PromoCode.objects.create(
            code="for_testing",
            amount=r
        )

    @property
    def random_amount(self):
        return random.randint(1, 100)

    @property
    def random_card_number(self):
        card_numbers = {
            "amex": "370000000000002",
            "discover": "6011000000000012",
            "visa": "4222222222222222",
            "mastercard": "5555555555554444",
        }
        return card_numbers.values()[random.randint(0, 3)]

    def get_base_random_data(self, **kwargs):
        now = datetime.datetime.now()
        data = {
            "first_name": self.random_donor_name,
            "last_name": self.random_donor_last_name,
            "amount": self.random_amount,
            "card_number": self.random_card_number,
            "ccv_code": "123",
            "expiration_month": "%02d" % now.month,
            "expiration_year": "%04d" % (now + datetime.timedelta(365)).year,
            "mailing_same_as_billing": u"1",
        }
        data.update(kwargs)
        return data

    def prefix_data(self, data, prefix=None):
        # TODO: Seems overly verbose, probably doing something wrong
        if not prefix:
            return data
        r = {}
        for k in data:
            r["%s-%s" % (prefix, k)] = data[k]
        return r

    def get_data_as_formset(self, data=None, prefix="form", total_forms=None,
            initial_forms=u"0", max_num_forms=u""):
        if data is None:
            data = []
        # TODO: write tests for this
        if type(data) is dict:
            data = [data, ]
        if not total_forms:
            total_forms = len(data)
        r = {
            "%s-TOTAL_FORMS" % prefix: total_forms,
            "%s-INITIAL_FORMS" % prefix: initial_forms,
            "%s-MAX_NUM_FORMS" % prefix: max_num_forms,
        }
        for idx, a in zip(range(len(data)), data):
            for k, v in a.items():
                r["%s-%d-%s" % (prefix, idx, k)] = v
        return r

    def get_payment_stub(self, successful=True, response_reason_text="Foobar"):
        fake_response = self.get_fake_purchase_response(successful,
                response_reason_text)
        fake = fudge.Fake()
        fake.provides("purchase") \
            .returns(fake_response)
        return fake

    def get_fake_purchase_response(self, successful=True,
            response_reason_text="Foobar"):
        fake = fudge.Fake()
        fake.has_attr(response_reason_text=response_reason_text)
        return {
            "status": "SUCCESS" if successful else "FAILURE",
            "response": fake,
        }

    def get_gateway_stub(self, payment_stub=None, successful=True,
            response_reason_text="Foobar"):
        if not payment_stub:
            payment_stub = self.get_payment_stub(successful=successful,
                    response_reason_text=response_reason_text)
        fake = fudge.Fake()
        fake.is_callable().returns(payment_stub)
        return fake

    def get_backend_stub(self, successful=True, reason="Foobar"):
        backend = fudge.Fake()
        backend.provides("get_form_class").returns(
                forms.CreditCardDonationForm)
        backend.provides("purchase").returns({
            "status": successful,
            "reason": reason,
            "response": "Foobar",
        })
        fake = fudge.Fake()
        fake.provides("get_backend").returns(backend)
        return fake

    @property
    def random_donation_and_form(self):
        donation = self.random_donation
        data = self.get_base_random_data(first_name=donation.donor.first_name,
                last_name=donation.donor.last_name, amount=donation.amount)
        donation_form = forms.AuthorizeDonationForm(data)
        donation_form.is_valid()
        return donation, donation_form



from decimal import Decimal
from django.contrib.auth.models import User
from django.contrib.localflavor.us import models as us
from django.db import models
from django.utils.translation import ugettext_lazy as _


class DonorAddress(models.Model):
    """Address associated with a ``Donor``"""
    address = models.CharField(max_length=255)
    city = models.CharField(max_length=20)
    state = us.USStateField()
    zipcode = models.CharField(max_length=10)

    def __unicode__(self):
        return "%(address)s, %(city)s, %(state)s, %(zipcode)s" % self.__dict__


class Donor(models.Model):
    """
    Represents a ``Donor`` associated with a ``Donation``

    This may or may not have an associated ``User`` with it, but
    should have one or more ``Donation`` associated with it.
    """
    user = models.ForeignKey(User, blank=True, null=True)
    first_name = models.CharField(max_length=250)
    last_name = models.CharField(max_length=250)
    address = models.ForeignKey(DonorAddress, related_name="addresses",
            null=True, blank=True)
    mailing_address = models.ForeignKey(DonorAddress,
            related_name="mailing_addresses", null=True, blank=True)
    # TODO: Make sure form widget is USPhoneNumberField
    phone = models.CharField(max_length=10, null=True, blank=True)
    email = models.EmailField(null=True, blank=True)

    def save(self, **kwargs):
        if self.user:
            if not self.first_name:
                self.first_name = self.user.first_name
            if not self.last_name:
                self.last_name = self.user.last_name
        super(Donor, self).save(**kwargs)

    def __unicode__(self):
        return "%s %s" % (self.first_name, self.last_name)


class DonationType(models.Model):
    """
    Provides a mechanism for creating types of donations

    Each ``DonationType`` has one or more ``DonationTypeOption`` which
    contains the actual data (amount, length, number of repeats, and so on).
    """
    name = models.CharField(
        max_length=100, help_text=_(u"Name of Donation Type")
    )

    def __unicode__(self):
        return self.name


class DonationTypeOption(models.Model):
    """Provides specifics for a particular option of a ``DonationType``"""
    donation_type = models.ForeignKey(DonationType, related_name="options")
    amount = models.PositiveIntegerField(help_text=_(u"Amount to donate"))
    length = models.PositiveIntegerField(default=1,
        help_text=_(u"Number of months per repeat "
                u"(1 is one month, 12 is one year)")
    )
    repeat = models.PositiveIntegerField(
        default=0, null=True, blank=True,
        help_text=_(u"Number of times (if any) this donation will repeat")
    )

    @property
    def name(self):
        return self.donation_type.name

    @property
    def is_repeating(self):
        return self.repeat > 0

    def __unicode__(self):
        return "%s (%d)" % (self.donation_type, self.amount)


class PromoCode(models.Model):
    """Codes that can be applied to a ``Donation`` to adjust its amount"""
    code = models.CharField(max_length=20, unique=True)
    amount = models.FloatField(
        help_text=_("Percent discount: 0 is no discount, 100 if free")
    )

    def __unicode__(self):
        return u"%s (%s%%)" % (self.code, self.amount)

    def calculate(self, donation):
        d = donation.amount * Decimal(1 - self.amount / 100.0)
        return Decimal(round(d, 2))


class Donation(models.Model):
    """Model representing the actual donation"""
    donor = models.ForeignKey(Donor)
    donation_type = models.ForeignKey(DonationTypeOption, null=True,
            blank=True)
    code = models.ForeignKey(PromoCode, null=True, blank=True)
    amount = models.DecimalField(max_digits=9, decimal_places=2)
    created = models.DateTimeField(auto_now_add=True)
    processed = models.BooleanField(default=False)
    attribution = models.CharField(max_length=255, default="")
    anonymous = models.BooleanField()

    def save(self, **kwargs):
        if self.donation_type and not self.amount:
            self.amount = self.donation_type.amount
        if self.code:
            self.amount = self.code.calculate(self)
        return super(Donation, self).save(**kwargs)

    def purchase(self, form):
        from . import backends
        return backends.get_backend().purchase(self, form)

    @property
    def is_repeating(self):
        return self.donation_type and self.donation_type.is_repeating or False

    def __unicode__(self):
        return "%s donated %s" % (self.donor, self.amount)



from copy import copy
from django import forms
from django.conf import settings
from django.contrib.auth.models import User

from .constants import MONTH_CHOICES
from .constants import YEAR_CHOICES
from .constants import MAILING_SAME_AS_BILLING

from . import models
from . import text

state_kwargs_fields = {}
if hasattr(settings, "ARMSTRONG_INITIAL_STATE"):
    state_kwargs_fields["initial"] = settings.ARMSTRONG_INITIAL_STATE


class BaseDonationForm(forms.Form):
    """
    Provides the basic fields common to all donation forms

    This is meant to be overridden by the forms returned by a
    ``Backend.get_form_class`` to provide specific implementations.
    """
    first_name = forms.CharField()
    last_name = forms.CharField()
    amount = forms.DecimalField(decimal_places=2)
    attribution = forms.CharField(required=False,
            help_text=text.get("donation.help_text.attribution"))
    anonymous = forms.BooleanField(required=False,
            label=text.get("donation.label.anonymous"))

    def __init__(self, data=None, prefix=None, *args, **kwargs):
        # TODO: provide custom prefixes to each sub-form
        self.mailing_same_as_billing = False
        # TODO: make this work with prefixes
        if (data and MAILING_SAME_AS_BILLING in data and
                data[MAILING_SAME_AS_BILLING]):
            self.mailing_same_as_billing = True
        self.donor_form = self.get_donor_form(data=data, prefix=prefix,
                *args, **kwargs)
        billing_address_prefix = "%sbilling" % ("" if not prefix else prefix)
        self.billing_address_form = self.get_billing_address_form(
                data=data, prefix=billing_address_prefix, empty_permitted=True,
                *args, **kwargs)
        self.billing_address_form.empty_permitted = True
        mailing_address_prefix = "%smailing" % ("" if not prefix else prefix)
        self.mailing_address_form = self.get_mailing_address_form(
                data=data, prefix=mailing_address_prefix, *args, **kwargs)
        self.mailing_address_form.empty_permitted = True
        super(BaseDonationForm, self).__init__(data=data, prefix=prefix,
                *args, **kwargs)

    def get_billing_address_form(self, *args, **kwargs):
        return DonorAddressForm(*args, **kwargs)

    def get_mailing_address_form(self, *args, **kwargs):
        return DonorAddressForm(*args, **kwargs)

    def get_donor_form(self, *args, **kwargs):
        return DonorForm(*args, **kwargs)

    def get_donation_kwargs(self):
        if not self.is_valid(donation_only=True):
            return {}
        return {
            "amount": self.cleaned_data["amount"],
            "attribution": self.cleaned_data["attribution"],
            "anonymous": bool(self.cleaned_data["anonymous"]),
        }

    def is_valid(self, donation_only=False):
        parent = super(BaseDonationForm, self)
        donation_is_valid = parent.is_valid()
        if not donation_is_valid and "amount" in self.errors:
            donation_type_field = self.add_prefix("donation_type_pk")
            if donation_type_field in self.data:
                donation_type_pk = self.data[donation_type_field]
                try:
                    dt = models.DonationTypeOption.objects.get(pk=donation_type_pk)

                    # We've made it this far, so create a copy of the data (which is
                    # most likely an immutable `QueryDict`) and adjust the amount to
                    # the correct value before re-running is_valid().
                    self.data = copy(self.data)
                    self.data["amount"] = dt.amount
                    self._errors = None
                    donation_is_valid = parent.is_valid()
                except models.DonationTypeOption.DoesNotExist:
                    donation_is_valid = False

        if donation_only:
            return donation_is_valid
        mailing_address_validity = self.billing_address_form.is_valid() \
                if self.mailing_same_as_billing \
                else self.mailing_address_form.is_valid()
        return all([
            donation_is_valid,
            self.donor_form.is_valid(),
            self.billing_address_form.is_valid(),
            mailing_address_validity,
        ])

    # TODO: support commit=True?
    def save(self, **kwargs):
        donation = models.Donation(**self.get_donation_kwargs())
        promo_code_field_name = self.add_prefix("promo_code")
        if (promo_code_field_name in self.data
                and self.data[promo_code_field_name]):
            donation.code = models.PromoCode.objects.get(
                    code=self.data[promo_code_field_name])
        if self.add_prefix("donation_type_pk") in self.data:
            donation.donation_type = models.DonationTypeOption.objects.get(
                    pk=self.data[self.add_prefix("donation_type_pk")])
        donor = self.donor_form.save(commit=False)
        try:
            user = User.objects.get(pk=self.data[self.add_prefix("user_pk")])
            donor.user = user
            if not donor.email:
                donor.email = user.email
        except (KeyError, ValueError):
            pass
        if self.billing_address_form.is_valid():
            donor.address = self.billing_address_form.save()
            donor.mailing_address = self.mailing_address_form.save() if \
                    not self.mailing_same_as_billing else donor.address
        donor.save()
        donation.donor = donor
        donation.save()
        return donation


class StripSensitiveFields(object):
    """
    Mixin for stripping sensitive information from an invalid form

    This is meant to be used by a ``Form`` object and strips the fields
    listed in ``fields_to_strip`` if ``is_valid`` fails.
    """
    fields_to_strip = []

    def is_valid(self, *args, **kwargs):
        r = super(StripSensitiveFields, self).is_valid(*args, **kwargs)
        if not r and self.fields_to_strip:
            empty_values = [""] * len(self.fields_to_strip)
            new_data = dict(zip(self.fields_to_strip, empty_values))
            self.data = copy(self.data)
            self.data.update(new_data)
        return r


class CreditCardDonationForm(StripSensitiveFields, BaseDonationForm):
    """
    Form representing the credit card donation process for Authorize.net

    It's possible that this can work with other backends, but it has not
    been tested.

    Each specific Donation form should create what is necessary for their
    backend's implemenation of `purchase`.

    .. todo:: Add widget that is smart for expiration dates
    """
    card_number = forms.CharField()
    ccv_code = forms.CharField()
    expiration_month = forms.ChoiceField(choices=MONTH_CHOICES)
    expiration_year = forms.ChoiceField(choices=YEAR_CHOICES)

    fields_to_strip = ["card_number", "ccv_code", ]

    def get_data_for_charge(self, donation, **kwargs):
        """
        Returns the data for charges

        This is used to build a dictionary of data that can not be
        retrieved directly from the ``Donation`` model.
        """
        raise NotImplementedError


class AuthorizeDonationForm(CreditCardDonationForm):
    """
    Represents a ``CreditCardDonationForm`` form for Authorize.net

    This form is returned by the ``AuthorizeNetBackend.get_form_class``
    and should not be accessed directly.
    """
    def get_data_for_charge(self, donation, recurring=False):
        """Returns the data for charges"""
        self.is_valid()
        card_number = "card_num" if not recurring else "card_number"
        data = {
            card_number: self.cleaned_data["card_number"],
            "card_code": self.cleaned_data["ccv_code"],
        }

        if recurring:
            data["expiration_date"] = u"%04d-%02d" % (
                    int(self.cleaned_data["expiration_year"]),
                    int(self.cleaned_data["expiration_month"]))
        else:
            data["exp_date"] = u"%02d-%04d" % (
                    int(self.cleaned_data["expiration_month"]),
                    int(self.cleaned_data["expiration_year"]))
        return data


class DonorForm(forms.ModelForm):
    """Simple ``ModelForm`` for the ``Donor`` model"""
    class Meta:
        model = models.Donor
        excludes = ("address", "mailing_address", )


class DonorAddressForm(forms.ModelForm):
    """Simple ``ModelForm`` for the ``DonorAddress`` model"""
    address = forms.CharField(widget=forms.Textarea)

    class Meta:
        model = models.DonorAddress
